from typing import Any, Protocol, Self, override, runtime_checkable

__all__ = ["PartialOrd", "Ord"]


@runtime_checkable
class PartialOrd(Protocol):
    def __lt__(self, other: Self, /) -> bool: ...


@runtime_checkable
class Ord(Protocol):
    def __lt__(self, other: Self, /) -> bool: ...
    @override
    def __eq__(self, other: Any, /) -> bool: ...



from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Generator, Hashable, Iterable, Literal, Sized, override

from serox.common import False_, True_
from serox.conftest import TESTING
from serox.convert import Into
from serox.default import Default
from serox.fmt import Debug
from serox.iter import Extend, FromIterator, IntoIterator, IntoParIterator, Iterator
from serox.misc import Clone
from serox.option import Null, Option, Some, is_null

__all__ = ["HashMap", "Entry"]

type Entry[K, V] = tuple[K, V]


class HashMap[K: Hashable, V](
    Default,
    Extend[Entry[K, V]],
    FromIterator[Entry[K, V]],
    IntoIterator[Entry[K, V]],
    IntoParIterator[Entry[K, V]],
    Into[dict[K, V]],
    Clone,
    Debug,
    Sized,
):
    def __init__(self, *args: Entry[K, V]) -> None:
        super().__init__()
        self.inner = {k: v for (k, v) in args}

    @override
    @classmethod
    def default(cls) -> HashMap[K, V]:
        return HashMap[K, V]()

    @classmethod
    def new(cls) -> HashMap[K, V]:
        return HashMap[K, V].default()

    @override
    def into(self) -> dict[K, V]:
        return self.inner

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({ repr(self.inner) })"

    def clear(self) -> None:
        self.inner.clear()

    def get(self, key: K, /) -> Option[V]:
        return Some[V].from_(self.inner.get(key))

    def entry(self, key: K, /) -> Option[Entry[K, V]]:
        return self.get(key).map(lambda v: (key, v))

    def insert(self, key: K, value: V) -> Option[V]:
        if is_null(opt := self.get(key)):
            self.inner[key] = value
        return opt

    def or_insert(self, default_key: K, default_value: V) -> Entry[K, V]:
        return (default_key, self.inner.setdefault(default_key, default_value))

    def contains_key(self, key: K) -> bool:
        return self.inner.__contains__(key)

    @override
    def __eq__(self, other: Any) -> bool:
        match other:
            case HashMap():
                return self.inner == other.inner  # pyright: ignore[reportUnknownMemberType]
            case _:
                return False

    def len(self) -> int:
        return self.inner.__len__()

    @override
    def __len__(self) -> int:
        return self.len()

    def is_empty(self) -> bool:
        return self.len() == 0

    @override
    def extend(self, iter: Iterable[Entry[K, V]], /) -> None:
        self.inner |= dict(iter)

    def remove(self, key: K, /) -> Option[V]:
        return Some[V].from_(self.inner.pop(key, None))

    def remove_entry(self, key: K, /) -> Option[Entry[K, V]]:
        return self.remove(key).map(lambda v: (key, v))

    @override
    @classmethod
    def from_iter(cls, iter: Iterable[Entry[K, V]], /) -> HashMap[K, V]:
        return HashMap(*iter)

    @override
    def iter(self) -> Entries[K, V, Literal[False]]:
        return Entries(self, par=False)

    def __iter__(self) -> Generator[Entry[K, V], None, None]:
        yield from self.iter()

    @override
    def par_iter(self) -> Entries[K, V, Literal[True]]:
        return Entries(self, par=True)

    def keys(self) -> Keys[K, Literal[False]]:
        return Keys[K, Literal[False]](self, par=False)

    def values(self) -> Values[V, Literal[False]]:
        return Values[V, Literal[False]](self, par=False)

    @override
    def clone(self) -> HashMap[K, V]:
        return HashMap[K, V](*self.inner.copy().items())


@dataclass(repr=True, init=False)
class Keys[K, P: (True_, False_)](Iterator[K, P]):
    def __init__(self, inner: HashMap[K, Any], /, par: P) -> None:
        super().__init__()
        self.iter = iter(inner.inner.keys())
        self.par = par

    @override
    def next(self) -> Option[K]:
        try:
            return Some(next(self.iter))
        except StopIteration:
            return Null()


@dataclass(repr=True, init=False)
class Values[V, P: (True_, False_)](Iterator[V, P]):
    def __init__(self, inner: HashMap[Any, V], /, par: P) -> None:
        super().__init__()
        self.iter = iter(inner.inner.values())
        self.par = par

    @override
    def next(self) -> Option[V]:
        try:
            return Some(self.iter.__next__())
        except StopIteration:
            return Null()


@dataclass(repr=True, init=False)
class Entries[K, V, P: (True_, False_)](Iterator[Entry[K, V], P]):
    def __init__(self, inner: HashMap[K, V], /, par: P) -> None:
        super().__init__()
        self.iter = iter(inner.inner.items())
        self.par = par

    @override
    def next(self) -> Option[Entry[K, V]]:
        try:
            return Some(next(self.iter))
        except StopIteration:
            return Null()


if TESTING:

    def test_hash_map():
        hm = HashMap(("foo", 0), ("bar", 1))
        _ = hm.iter().collect(HashMap[str, int])
        _ = hm.keys()
        _ = hm.values()
        print(hm)



from __future__ import annotations  # noqa: I001
from dataclasses import dataclass
from random import Random as Rng
from .common import True_, False_
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Iterable,
    Self,
    overload,
    override,
)

from serox.convert import Into
from serox.fmt import Debug
from serox.question_mark import qmark
from serox.result import ErrShortCircuit, Ok, Result, Err
from .iter import (
    DoubleEndedIterator,
    FromIterator,
    IntoIterator,
    IntoParIterator,
    Iterator,
    Extend,
)
from .cmp import PartialOrd
from .default import Default


from serox.conftest import TESTING

from serox.misc import Clone, IndexType, SizedIndexable
from serox.option import Null, Option, Some

if TYPE_CHECKING:
    from serox import Range

__all__ = [
    "Vec",
]

type Fn0[T] = Callable[[], T]
type Fn1[T, U] = Callable[[T], U]


@dataclass(repr=True, frozen=True, kw_only=True)
class Iter[Item, P: (True_, False_)](DoubleEndedIterator[Item, P]):
    data: SizedIndexable[Item]
    par: P
    end_or_len: int
    ptr: int

    """
    Wrapper around a native Python :class:`list` to endow it with `Vec`-like functionality.
    """

    @classmethod
    def new[Item2, Par2: (True_, False_)](
        cls, data: SizedIndexable[Item2], /, par: Par2 = False
    ) -> Iter[Item2, Par2]:
        return Iter(
            data=data,
            par=par,
            ptr=0,
            end_or_len=len(data),
        )

    def _next(self, back: bool) -> Option[Item]:
        if self.ptr < self.end_or_len:
            ptr = -self.ptr if back else self.ptr
            item = Some(self.data[ptr])
            object.__setattr__(self, "ptr", self.ptr + 1)
            return item
        return Null()

    @override
    def next(self) -> Option[Item]:
        return self._next(back=False)

    @override
    def next_back(self) -> Option[Item]:
        return self._next(back=True)


@dataclass(repr=True, frozen=True, kw_only=True)
class SampledIndexable[Item](SizedIndexable[Item]):
    data: SizedIndexable[Item]
    indices: list[IndexType]

    @classmethod
    def new[Item2](
        cls,
        data: SizedIndexable[Item2],
        indices: list[IndexType],
    ) -> SampledIndexable[Item2]:
        return SampledIndexable(data=data, indices=indices)

    @override
    def __getitem__(self, index: IndexType, /) -> Item:
        return self.data[self.indices[index]]

    @override
    def __len__(self) -> int:
        return len(self.indices)


class Vec[T](
    IntoIterator[T],
    IntoParIterator[T],
    SizedIndexable[T],
    FromIterator[T],
    Default,
    Debug,
    Clone,
    Extend[T],
    Into[list[T]],
):
    """
    Wrapper around a native Python :class:`list` to endow it with `Vec`-like functionality.
    """

    def __init__(self, *args: T) -> None:
        super().__init__()
        self.inner = list(args)

    @override
    @classmethod
    def default(cls: type[Vec[T]]) -> Vec[T]:
        """
        Returns a new, empty `Vec` as the default.

        :returns: A new, empty `Vec`.
        """
        return Vec[T]()

    @classmethod
    def new(cls) -> Vec[T]:
        """
        Returns a new, empty `Vec`.

        :returns: A new, empty `Vec`.
        """
        return Vec[T].default()

    @override
    def into(self) -> list[T]:
        return self.inner

    @override
    def iter(self) -> DoubleEndedIterator[T, False_]:
        """
        Returns an iterator over the underlying list.

        The iterator yields all items from start to end.

        :returns: A double-ended iterator over the underlying list.
        """
        return Iter.new(self, par=False)

    @override
    def par_iter(self) -> DoubleEndedIterator[T, True_]:
        """
        Returns a parallel iterator over the underlying list.

        The iterator yields all items from start to end.

        :returns: A parallel double-ended iterator over the underlying list.
        """
        return Iter.new(self, par=True)

    def __iter__(self) -> Generator[T, None, None]:
        """
        Returns a generator over the elements of the `Vec`.

        :yield: Elements of the `Vec` in order.
        """
        yield from self.iter()

    @override
    def __eq__(self, other: Any) -> bool:
        match other:
            case Vec():
                return self.inner == other.inner  # pyright: ignore[reportUnknownMemberType]
            case _:
                return False

    @override
    def __repr__(self) -> str:
        return f"Vec({self.inner.__repr__()[1:-1]})"

    @override
    def __getitem__(self, index: int, /) -> T:
        return self.inner[index]

    @overload
    def get(self, index: Range[Any], /) -> Option[Vec[T]]: ...
    @overload
    def get(self, index: int, /) -> Option[T]: ...
    def get(self, index: int | Range[Any], /) -> Option[T] | Option[Vec[T]]:
        """
        Returns an element or sub-vector depending on the type of index.

        - If given a position, returns the element at that position or `Null` if out of bounds.
        - If given a range, returns the sub-vector corresponding to that range, or `Null` if out of
        bounds.
        """
        from serox import Range

        match index:
            case Range():
                if index.contains(self.len()):
                    return Some(Vec(*self.inner[index.start : index.end]))
                return Null()
            case int():
                if index >= self.len():
                    return Null()
                return Some(self.inner[index])

    @override
    def __len__(self) -> int:
        return len(self.inner)

    def len(self) -> int:
        """
        Returns the number of elements in the vector, also referred to as its 'length'.
        """
        return self.__len__()

    def is_empty(self) -> bool:
        """
        :returns: `True` if the vector contains no elements.
        """
        return self.len() == 0

    def split_off(self, at: int, /) -> Vec[T]:
        """
        Splits the vector into two at the given index, retaining the head :math:`[0, at)` and
        returning the tail :math:`[at. len)`.

        :param at: The index to split the vector at.
        :returns: A newly-allocated vector containing the elements :math:`[at, len)` of the tail.
        """
        tail = self.inner[at:]
        self.inner = self.inner[:at]
        return Vec(*tail)

    @classmethod
    def full(cls, value: T, n: int) -> Vec[T]:
        """
        Initializes a new `Vec` with `n` copies of `value`.

        :param value: The value to fill the `Vec` with.
        :param n: The number of copies of `value` to fill the `Vec` with.
        :returns: A new `Vec` with `n` copies of `value`.
        """
        return Vec[T](*(value for _ in range(n)))

    def push(self, item: T, /) -> None:
        """
        Appends an item to the end of the `Vec`.

        :param item: The item to append.
        """
        self.inner.append(item)

    def clear(self) -> None:
        """
        Removes all elements from the `Vec`.
        """
        self.inner.clear()

    @override
    def extend(self, items: Iterable[T], /) -> None:
        """
        Extends the `Vec` with the elements from the given iterable.

        :param items: The iterable of items to extend the `Vec` with.
        """
        self.inner.extend(items)

    def append(self, other: Self, /):
        """
        Appends the elements of another `Vec` to this `Vec`, then clears the other `Vec`.

        :param other: The `Vec` to append.
        """
        self.inner.extend(other.iter())
        other.clear()

    def first(self) -> Option[T]:
        """
        Returns the first element of the `Vec`, or `Null` if the `Vec` is empty.

        :returns: The first element of the `Vec`, or `Null` if the `Vec` is empty.
        """
        return Null() if self.is_empty() else Some(self[0])

    def last(self) -> Option[T]:
        """
        Returns the last element of the `Vec`, or `Null` if the `Vec` is empty.

        :returns: The last element of the `Vec`, or `Null` if the `Vec` is empty.
        """
        return Null() if self.is_empty() else Some(self[-1])

    def sort_by(
        self: Vec[T],
        compare: Fn1[T, PartialOrd],
        /,
        reverse: bool = False,
    ) -> None:
        """
        Sorts the `Vec` in place using the given comparison function.

        :param compare: The comparison function to use for sorting.
        :param reverse: If `True`, the `Vec` is sorted in reverse order.
        """
        self.inner.sort(key=compare, reverse=reverse)

    def sort[U: PartialOrd](
        self: Vec[U],
        reverse: bool = False,
    ) -> None:
        """
        Sorts the `Vec` in place.

        :param reverse: If `True`, the `Vec` is sorted in reverse order.
        """
        self.inner.sort(reverse=reverse)

    def pop(self) -> Option[T]:
        """
        Pop the last element from the vector and return it.
        :returns: `Some(x)`, where `x` is the last element of the vector if it is not empty,
            otherwise `Null`.
        """
        if self.is_empty():
            return Null()
        return Some(self.inner.pop())

    def remove(self, index: int) -> T:
        """
        Removes and returns the element at position `index` within the vector, shifting all
        elements after it to the left.

        :param index: The index of the element to remove.
        :returns: The removed element previously at index `index` of the vector.
        :raises IndexError: if `index` is out of bounds.
        """
        if index >= self.len():
            raise IndexError(f"removal index (is {index}) should be < len (is {self.len()})")
        return self.inner.pop(index)

    def insert(self, index: int, element: T) -> None:
        """
        Inserts an element at position `index` within the vector, shift all elements after it to
        the right.

        :param index: The index at which to insert the new `element`.
        :param element: The new element to insert at index `index`.
        """
        self.inner.insert(index, element)

    @override
    @classmethod
    def from_iter[P: (True_, False_)](cls, s: Iterator[T, P], /) -> Vec[T]:
        """
        Creates a new `Vec` from an iterator.

        :param s: The iterator to create the `Vec` from.
        :returns: A new `Vec` containing the elements from the iterator.
        """
        return Vec(*s)

    @override
    def clone(self) -> Self:
        """
        Creates a clone of the `Vec`.

        :returns: A clone of the `Vec`.
        """
        return super().clone()

    def retain(self, f: Fn1[T, bool], /) -> None:
        """
        Retains only the elements specified by the predicate function.

        :param f: The predicate function to determine which elements to retain.
        """
        self.inner = list(filter(f, self.inner))

    def choose(self, rng: Rng) -> Option[T]:
        """
        Emulates `SliceRandom::choose` from the `rand` crate.

        Randomly selects an element from the vector and returns it.
        This returns :class:`~Null` if the vector is empty.

        :param rng: The random number generator to use to choose the element.
        :returns: `Some(x)`, where `x` is the randomly-chosen element if the vector is not empty,
            otherwise `Null`.
        """
        if self.is_empty():
            return Null()
        return Some(rng.choice(self.inner))

    def choose_multiple(self, rng: Rng, amount: int) -> Iter[T, False_]:
        """
        Emulates `SliceRandom::choose_multiple` from the `rand` crate.

        Randomly selects `amount` elements - or all elements if `amount` exceeds the
        vector's length - from the vector without replacement and returns an iterator over them.

        :param rng: The random number generator to use to sample the elements without replacement.
        :param amount: The number of elements to sample without replacement,
            capped at the length of the vector.
        :returns: An iterator over the chosen elements.
        """
        # cap the sample size at the population size
        amount = min(amount, self.len())
        # sample without replacement
        sampled_indices = rng.sample(range(self.len()), k=amount)
        return Iter.new(SampledIndexable.new(self, sampled_indices), par=False)

    @qmark
    def choose_multiple_weighted(
        self,
        rng: Rng,
        amount: int,
        weight: Fn1[T, float],
    ) -> Result[Iter[T, False_], ValueError]:
        """
        Similar to :meth:`~Vec.choose_multiple`, but where the likelihood of each element’s
        inclusion in the output may be specified. The elements are returned in an arbitrary,
        unspecified order.

        :param rng: The random number generator to use to sample the elements without replacement.

        :param amount: The number of elements to sample without replacement,
            capped at the length of the vector.

        :param weight: Weighting function mapping each item :math:`x` to a relative likelihood
            :math:`weight(x)`. The probability of each item being selected is thus
            :math:`weight(x) / s`, where :math:`s` is the sum of all :math:`weight(x)`.

        :returns: An iterator over the chosen elements.
        """

        def call(x: T) -> float:
            match weight(x):
                case p if (0 <= p) and math.isfinite(p):
                    return p
                case _:
                    raise ErrShortCircuit(ValueError("Invalid weight"))

        weights = self.iter().map(call).collect(Vec[float]).into()
        # cap the sample size at the population size
        amount = min(amount, self.len())
        return Ok(Iter.new(rng.choices(self.inner, weights=weights, k=amount), par=False))

    def shuffle(self, rng: Rng) -> None:
        """
        Shuffles the elements of the `Vec` in place using the given random number generator.

        :param rng: The random number generator to use for shuffling.
        """
        rng.shuffle(self.inner)

    def fill(self, value: T, /) -> None:
        """
        Fills the `Vec` with the given value.

        :param value: The value to fill the `Vec` with.
        """
        self.inner = self.len() * [value]

    def fill_with(self, f: Fn0[T], /) -> None:
        """
        Fills the `Vec` with values generated by the given function.

        :param f: The function to generate values to fill the `Vec` with.
        """
        self.inner = [f() for _ in range(self.len())]

    def join[U: str](self: Vec[U], separator: U, /) -> U:
        """
        Joins the elements of the `Vec` into a string, separated by the given separator.

        :param separator: The separator to use between elements.
        :returns: A string with the elements of the `Vec` joined by `separator`.
        """
        return separator.join(self.inner)

    def repeat(self, n: int, /) -> Vec[T]:
        """
        Repeats the elements of the `Vec` `n` times.

        :param n: The number of times to repeat the elements.
        :returns: A new `Vec` with the elements repeated `n` times.
        """
        return Vec(*(self.inner * n))

    def flatten[U](self: Vec[Vec[U]]) -> Vec[U]:
        """
        Flattens a `Vec` of `Vec`s into a single `Vec`.

        :returns: A new `Vec` with the elements of the inner `Vec`s.
        """
        return Vec(*(inner for outer in self for inner in outer))

    def dedup(self) -> None:
        """
        Removes consecutive duplicate elements from the `Vec`.

        This method retains only the first occurrence of each group of consecutive duplicate elements.

        :returns: None
        """
        # iterator to start from the second (first) element
        next_iter = self.iter()
        # offset the iterator by one. If there is no second element, return `Null`, otherwise
        # proceed with the deduplication algorithm.
        match next_iter.advance_by(1):
            case Ok(_):

                def not_repeated(pair: tuple[T, T], /) -> bool:
                    return pair[0] != pair[1]

                # iterator starting from the first (zeroth) element
                prev_iter = self.iter()
                self.inner = (
                    prev_iter.zip_longest(next_iter)
                    # retain paris with unequal elements
                    .filter(not_repeated)
                    # flatten by taking the first element of each successive pair.
                    .map(lambda x: x[0])
                    .collect(Vec[T])
                    .into()
                )
            case Err(_):
                return None


if TESTING:
    import math

    def test_sort():
        vec = Vec[int](3, 2, 1)
        vec.sort()
        assert vec == Vec(1, 2, 3)
        vec.sort(reverse=True)
        assert vec == Vec(3, 2, 1)
        vec.sort_by(lambda x: -x, reverse=True)
        assert vec == Vec(1, 2, 3)

    def test_filter():
        vec = Vec[int](1, 2, 3, 6)
        evens = vec.iter().filter(lambda x: x % 2 == 0).collect(Vec[int])
        assert evens == Vec(2, 6)

    def test_chain():
        vec = Vec[int](1, 2, 3)
        vec2 = Vec[int](2, 3, 4)
        res: Vec[int] = vec.iter().chain(vec2.iter()).collect(Vec)
        assert res == Vec(1, 2, 3, 2, 3, 4)

    def test_take():
        vec = Vec[float].full(math.pi, 3)
        taken = vec.iter().take(2).collect(Vec[float])
        assert len(taken) == 2
        taken = vec.iter().take(5).collect(Vec[float])
        assert len(taken) == vec.len()

    def test_take_while():
        vec = Vec[float](*range(5))
        lt5 = vec.iter().take_while(lambda x: x < 5).collect(Vec[float])
        assert len(lt5) == vec.len()
        lt3 = vec.iter().take_while(lambda x: x < 3).collect(Vec[float])
        assert len(lt3) == 3
        vec2 = Vec[Vec[int]](Vec(1, 2, 3))
        ls = [2, 3, 4]
        vec.extend(ls)
        _ = vec2.clone()

    def test_par_iter():
        vec = Vec[float].full(3.14, n=3)
        res = vec.par_iter().take(2).collect(Vec[float])
        assert len(res) == 2

    def test_agg():
        vec = Vec[float](1.0, 2.0, 3.0, -3.0)
        sum_ = vec.iter().sum()
        assert sum_ == 3.0
        prod = vec.iter().product()
        assert prod == -18.0
        max_ = vec.iter().max()
        assert max_ == 3.0

    def test_chunking():
        vec = Vec(*range(97))
        chunks = vec.iter().array_chunk(10)
        assert chunks.count() == math.ceil(100 / 10)
        chunks = vec.par_iter().array_chunk(10)
        assert chunks.count() == math.ceil(100 / 10)

    def test_rng():
        vec = Vec(0, 1, 2, 5, 7, 12, 19)
        rng = Rng(42)
        _ = vec.choose(rng)
        chosen = vec.choose_multiple(rng, 4)
        assert chosen.count() == 4
        chosen = vec.choose_multiple_weighted(rng, 4, lambda _: 1 / vec.len())
        assert chosen.unwrap().count() == 4
        # should short circuit upon violation of the non-negativity constraint
        assert vec.choose_multiple_weighted(rng, 4, lambda _: -1).is_err()

    def test_dedup():
        vec = Vec(1, 1, 3, 2, 2, 2, 4, 5, 4, 4)
        vec.dedup()
        assert vec == Vec(1, 3, 2, 4, 5, 4)

    def test_join():
        vec = Vec("foo", "bar", "baz")
        assert vec.join(",") == "foo,bar,baz"



from .collections import *
from .io import *
from .iter import *
from .misc import *
from .option import *
from .optional import *
from .question_mark import *
from .range import *
from .result import *
from .vec import *



from typing import Protocol, Self

__all__ = [
    "From",
    "Into",
]


class From[T](Protocol):
    @classmethod
    def from_(cls, s: T, /) -> Self: ...


class Into[T](Protocol):
    def into(self) -> T: ...



from __future__ import annotations
import copy
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Callable,
    Generator,
    Hashable,
    NoReturn,
    Protocol,
    TypeGuard,
    cast,
    final,
    overload,
    override,
)

from serox.conftest import TESTING
from serox.convert import Into
from serox.misc import Clone

from .fmt import Debug

if TYPE_CHECKING:
    from serox.option import Option

__all__ = [
    "Ok",
    "Err",
    "Result",
    "is_ok",
    "is_err",
]


@final
class UnwrapFailed(Exception):
    @override
    def __init__(self, msg: str, error: Debug, /) -> None:
        super().__init__(f"{msg}: {error.__repr__}")


type Fn0[T] = Callable[[], T]
type Fn1[T, U] = Callable[[T], U]


class _Result[T, E](
    Clone,
    Hashable,
    Protocol,
):
    def __iter__(self: Result[T, E]) -> Generator[T, None, None]:
        match self:
            case Ok(x):
                yield x
            case Err():
                ...

    def __next__(self: Result[T, E]) -> T:
        match self:
            case Ok(x):
                return x
            case Err(_):
                raise StopIteration

    # Property emulative of Rust's `?` operator when
    # used in conjunction with [`serox::question_mark::qmark`].
    @property
    def q(self: Result[T, E]) -> T:
        match self:
            case Ok(t):
                return t
            case Err(err):
                raise ErrShortCircuit(err)

    def __invert__(self: Result[T, E]) -> T:
        return self.q

    def unwrap(self: Result[T, E]) -> T:
        match self:
            case Ok(x):
                return x
            case Err(err):
                raise UnwrapFailed("called `Result.unwrap()` on an `Err` value", err)

    def unwrap_or(self: Result[T, E], default: T) -> T:
        match self:
            case Ok(x):
                return x
            case Err(_):
                return default

    def unwrap_or_else(self: Result[T, E], f: Fn0[T]) -> T:
        match self:
            case Ok(x):
                return x
            case Err(_):
                return f()

    @overload
    def expect(self: Ok[T, E], msg: str) -> T: ...
    @overload
    def expect(self: Err[T, E], msg: str) -> NoReturn: ...
    def expect(self: Result[T, E], msg: str) -> T | NoReturn:
        match self:
            case Ok(t):
                return t
            case Err(e):
                raise UnwrapFailed(msg, e)

    @overload
    def expect_err(self: Ok[T, E], msg: str) -> NoReturn: ...
    @overload
    def expect_err(self: Err[T, E], msg: str) -> E: ...
    def expect_err(self: Result[T, E], msg: str) -> E | NoReturn:
        match self:
            case Ok(t):
                raise UnwrapFailed(msg, t)
            case Err(e):
                return e

    @overload
    def map[U](self: Err[T, E], f: Fn1[T, U], /) -> Err[U, E]: ...
    @overload
    def map[U](self: Ok[T, E], f: Fn1[T, U], /) -> Ok[U, E]: ...
    def map[U](self: Result[T, E], f: Fn1[T, U], /) -> Result[U, E]:
        match self:
            case Ok(t):
                return Ok(f(t))
            case Err(e):
                return Err[U, E](e)

    @overload
    def map_err[F](self: Err[T, E], f: Fn1[T, F], /) -> Err[T, F]: ...
    @overload
    def map_err[F](self: Ok[T, E], f: Fn1[T, F], /) -> Ok[T, F]: ...
    def map_err[F](self: Result[T, E], f: Fn1[E, F], /) -> Result[T, F]:
        match self:
            case Ok(t):
                return Ok[T, F](t)
            case Err(e):
                return Err[T, F](f(e))

    def map_or[U](self: Result[T, E], default: U, f: Fn1[T, U], /) -> U:
        match self:
            case Ok(x):
                return f(x)
            case Err(_):
                return default

    def map_or_else[U](self: Result[T, E], default: Fn0[U], f: Fn1[T, U], /) -> U:
        match self:
            case Ok(x):
                return f(x)
            case Err(_):
                return default()

    def is_ok(self: Result[T, E]) -> bool:
        return isinstance(self, Ok)

    def is_err(self: Result[T, E]) -> bool:
        return not self.is_ok()

    def ok(self: Result[T, E]) -> Option[T]:
        from .option import Null, Some

        match self:
            case Ok(x):
                return Some(x)
            case Err(_):
                return Null[T]()

    def err(self: Result[T, E]) -> Option[E]:
        from .option import Null, Some

        match self:
            case Ok(_):
                return Null()
            case Err(x):
                return Some(x)

    @overload
    def and_then[U](self: Err[T, E], f: Fn1[T, U], /) -> Err[U, E]: ...
    @overload
    def and_then[U](self: Ok[T, E], f: Fn1[T, U], /) -> Result[U, E]: ...
    def and_then[U](self: Result[T, E], op: Callable[[T], Result[U, E]], /) -> Result[U, E]:
        match self:
            case Ok(t):
                return op(t)
            case Err(e):
                return Err[U, E](e)

    def or_(self: Result[T, E], res: Result[T, E], /) -> Result[T, E]:
        return self.__or__(res)

    def __or__(self: Result[T, E], res: Result[T, E], /) -> Result[T, E]:
        match self:
            case Ok(t):
                return Ok(t)
            case Err():
                return res

    def or_else[F](self: Result[T, E], op: Fn1[E, Result[T, F]], /) -> Result[T, F]:
        match self:
            case Ok(t):
                return Ok(t)
            case Err(e):
                return op(e)

    @override
    def clone(self: Result[T, E]) -> Result[T, E]:
        match self:
            case Ok(x):
                return Ok(copy.deepcopy(x))
            case Err(x):
                return Err(copy.deepcopy(x))

    def transpose(self: Result[Option[T], E]) -> Option[Result[T, E]]:
        from .option import Null, Some

        match self:
            case Ok(Some(x)):
                return Some(Ok(x))
            case Err(e):
                return Some(Err(e))
            case _:  # Ok(Null())
                return Null()


@final
@dataclass(
    eq=True,
    repr=True,
    frozen=True,
)
class Ok[T, E](_Result[T, E]):
    __match_args__ = ("value",)
    __slots__ = ("value",)

    value: T


@final
@dataclass(
    eq=True,
    frozen=True,
    repr=True,
)
class Err[T, E](_Result[T, E]):
    __match_args__ = ("value",)
    __slots__ = ("value",)

    value: E


type Result[T, E] = Ok[T, E] | Err[T, E]


# Applying [`TypeGuard`] to a `self` parameter is not permitted so
# we have to resort to defining external `is_ok` and `is_err` functions
# in order to generate boolean checkers capable of type narrowing.
def is_ok[T, E](x: Result[T, E], /) -> TypeGuard[Ok[T, E]]:
    return x.is_ok()


def is_err[T, E](x: Result[T, E], /) -> TypeGuard[Err[T, E]]:
    return x.is_err()


class ErrShortCircuit[T, E](Into[Err[T, E]], Exception):
    __match_args__ = ("err",)

    def __init__(self, err: E) -> None:
        super().__init__()
        self.err = err

    @override
    def into(self) -> Err[T, E]:
        return Err[T, E](self.err)


if TESTING:
    from typing import cast

    import pytest

    @pytest.fixture()
    def foo_res() -> Ok[str, Exception]:
        return Ok("foo")

    def test_matching(foo_res: Result[str, Exception]):
        match foo_res:
            case Ok(v):
                assert v == "foo"
            case Err(err):
                raise err
        option2 = cast(Result[str, Exception], Err("bar"))
        match option2:
            case Ok(v):
                raise ValueError
            case v:
                ...

    def test_map(foo_res: Result[str, Exception]):
        res = foo_res.map(lambda x: x + "bar")
        assert res.unwrap() == "foobar"
        foo_res = Err[str, Exception](Exception("Error"))
        res = foo_res.unwrap_or_else(lambda: "baz")
        assert res == "baz"

    def test_expect(foo_res: Result[str, Exception]):
        ok = foo_res.expect("Expected value 'foo'")
        assert ok == "foo"
        with pytest.raises(UnwrapFailed):
            _ = foo_res.expect_err("Expected error")
        err = Err[str, str]("Some error")
        with pytest.raises(UnwrapFailed):
            err.expect("Expected value 'foo'")

    def test_iter(foo_res: Result[str, Exception]):
        # Ok[T, _] should yield a single item of type T
        val = next(foo_res)
        assert val == "foo"
        val = list(iter(foo_res))
        assert len(val) == 1

        ls: list[str] = []
        ls.extend(foo_res)
        assert len(ls) == 1
        assert ls[0] == "foo"

        # Err[T, _] should yield nothing
        err = Err[str, str]("Error")
        val = list(iter(err))
        assert len(val) == 0



import functools
from typing import Callable, cast, overload

from serox.conftest import TESTING

from .option import Null, NullShortCircuit, Option, Some
from .result import Err, ErrShortCircuit, Ok, Result

__all__ = [
    "qmark",
]


@overload
def qmark[T, E, **P](
    f: Callable[P, Result[T, E]],
) -> Callable[P, Result[T, E]]: ...
@overload
def qmark[T, **P](
    f: Callable[P, Option[T]],
) -> Callable[P, Option[T]]: ...


def qmark[T, E, **P](
    f: Callable[P, Result[T, E]] | Callable[P, Option[T]],
) -> Callable[P, Result[T, E]] | Callable[P, Option[T]]:
    @functools.wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Result[T, E] | Option[T]:
        try:
            return f(*args, **kwargs)
        except ErrShortCircuit as err:  # pyright: ignore[reportUnknownVariableType]
            err = cast(ErrShortCircuit[T, E], err)
            return err.into()
        except NullShortCircuit:
            return Null[T]()

    return wrapper  # pyright: ignore[reportReturnType]


if TESTING:
    import pytest

    def test_qmark():
        @qmark
        def res_fn(s: Result[str, str]) -> Result[str, str]:
            return Ok(s.q)

        err = Err[str, str]("Error")
        assert res_fn(err) == err
        ok = Ok[str, str]("foo")
        assert res_fn(ok) == ok

        @qmark
        def opt_fn(s: Option[str]) -> Option[str]:
            return Some(s.q)

        null = Null[str]()
        assert opt_fn(null) == null
        some = Some("foo")
        assert opt_fn(some) == some

        def opt_fn_no_dec(s: Option[str]) -> Option[str]:
            return Some(s.q)

        null = Null[str]()
        with pytest.raises(NullShortCircuit):
            assert opt_fn_no_dec(null) == null

        @qmark  # pyright: ignore[reportCallIssue, reportArgumentType, reportUntypedFunctionDecorator]
        def opt_fn_bad_sig(_: str) -> str: ...

        del opt_fn_bad_sig



# pyright: reportImportCycles=none
from __future__ import annotations
from dataclasses import dataclass
import functools
import operator
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Iterable,
    Protocol,
    Self,
    cast,
    final,
    override,
    runtime_checkable,
)
from typing import Iterator as NativeIterator

from joblib import (  # pyright: ignore[reportMissingTypeStubs]
    Parallel,
    delayed,  # pyright: ignore[reportUnknownVariableType]
)

from serox.cmp import Ord
from serox.common import False_, True_
from serox.conftest import TESTING
from serox.misc import SelfAddable, SelfMultiplicable

if TYPE_CHECKING:
    from serox.option import Null, Option, Some
    from serox.result import Result


__all__ = [
    "ArrayChunk",
    "Bridge",
    "Chain",
    "Chunk",
    "DoubleEndedIterator",
    "Extend",
    "Filter",
    "FilterMap",
    "IntoIterator",
    "Iterator",
    "Map",
    "Rev",
    "Take",
    "TakeWhile",
    "Zip",
]


type Fn1[T, U] = Callable[[T], U]


class FromIterator[A](Protocol):
    @classmethod
    def from_iter[P: (True_, False_)](cls, iter: Iterator[A, P], /) -> Self: ...


def _identity[T](x: T) -> T:
    """
    Identity function that returns the input value.
    """
    return x


@runtime_checkable
class Iterator[Item, Par: (True_, False_)](Protocol):
    par: Par

    def next(self) -> Option[Item]: ...

    def __next__(self) -> Item:
        from serox.option import Null, Some

        match self.next():
            case Some(x):
                return x
            case Null():
                raise StopIteration

    @final
    def _iter(self) -> Generator[Item, None, None]:
        from serox.option import Null, Some

        while True:
            match self.next():
                case Some(v):
                    yield v
                case Null():
                    break

    @final
    def __iter__(self) -> Generator[Item, None, None]:
        if self.par:
            yield from cast(
                Generator[Item, None, None],
                Parallel(n_jobs=-1, verbose=0)(delayed(_identity)(i) for i in self._iter()),
            )
        else:
            yield from self._iter()

    @final
    def zip[U](self, other: Iterator[U, Par]) -> Zip[Item, U, Par]:
        return Zip(a=self, b=other, par=self.par)

    @final
    def zip_longest[U](self, other: Iterator[U, Par]) -> ZipLongest[Item, U, Par]:
        return ZipLongest(a=self, b=other, par=self.par)

    @final
    def chain(self, other: Iterator[Item, Par]) -> Chain[Item, Par]:
        return Chain(a=self, b=other, par=self.par)

    @final
    def map[B](self, f: Fn1[Item, B], /) -> Map[Item, B, Par]:
        return Map(self, f, par=self.par)

    @final
    def fold[B](self, init: B, f: Callable[[B, Item], B], /) -> B:
        from .option import is_some

        accum = init
        while is_some(v := self.next()):
            accum = f(accum, v.unwrap())
        return accum

    def count(self) -> int:
        def increment(count: int, _: Item) -> int:
            return count + 1

        return self.fold(0, increment)

    def array_chunk(self, n: int, /) -> ArrayChunk[Item, Par]:
        return ArrayChunk(self, n, par=self.par)

    def filter(self, f: Fn1[Item, bool]) -> Filter[Item, Par]:
        return Filter(self, f, par=self.par)

    def filter_map[B](self, f: Fn1[Item, Option[B]]) -> FilterMap[Item, B, Par]:
        return FilterMap(self, f, par=self.par)

    @final
    def collect[B: "FromIterator[Any]"](self, collection: type[B], /) -> B:
        return collection.from_iter(self)

    def nth(self, n: int) -> Option[Item]:
        from serox.option import Null, Some

        while n > 0:
            match self.next():
                case Some(_):
                    n -= 1
                case Null():
                    return Null()
        return self.next()

    def advance_by(self, n: int, /) -> Result[None, int]:
        from .option import Null, Some
        from .result import Err, Ok

        while n > 0:
            match self.next():
                case Some(_):
                    n -= 1
                case Null():
                    return Err(n)
        return Ok(None)

    def for_each(self, f: Fn1[Item, None]) -> None:
        def call(_: None, item: Item, /) -> None:
            return f(item)

        self.fold(None, call)

    def take(self, n: int, /) -> Take[Item, Par]:
        return Take(self, n, par=self.par)

    def take_while(self, f: Fn1[Item, bool], /) -> TakeWhile[Item, Par]:
        return TakeWhile(self, f, par=self.par)

    def sum[U: SelfAddable](self: Iterator[U, Par]) -> U:
        return functools.reduce(operator.add, self)

    def product[U: SelfMultiplicable](self: Iterator[U, Par]) -> U:
        return functools.reduce(operator.mul, self)

    def all(self, f: Fn1[Item, bool], /) -> bool:
        return all(f(item) for item in self)

    def any(self, f: Fn1[Item, bool], /) -> bool:
        return any(f(item) for item in self)

    def max[U: Ord](self: Iterator[U, Par]) -> U:
        return max(self)

    def min[U: Ord](self: Iterator[U, Par]) -> U:
        return min(self)

    def par_bridge(self) -> Iterator[Item, True_]:
        object.__setattr__(self, "par", True)
        return cast(Iterator[Item, True_], self)


class Chunk[Item](list[Item], FromIterator[Item]):
    @override
    @classmethod
    def from_iter(cls, iter: Iterable[Item], /) -> Chunk[Item]:
        return Chunk(iter)

    def len(self) -> int:
        return self.__len__()

    def is_empty(self) -> bool:
        return self.len() == 0


@dataclass
class ArrayChunk[Item, P: (True_, False_)](Iterator[Chunk[Item], P]):
    iter: Iterator[Item, P]
    n: int
    par: P

    @override
    def next(self) -> Option[Chunk[Item]]:
        from .option import Null, Some

        match self.iter.take(self.n).collect(Chunk[Item]):
            case []:
                return Null()
            case chunk:
                return Some(chunk)


class IntoIterator[T](Protocol):
    def iter(self) -> Iterator[T, False_]: ...


class IntoParIterator[T](Protocol):
    def par_iter(self) -> Iterator[T, True_]: ...


class DoubleEndedIterator[Item, P: (True_, False_)](Iterator[Item, P], Protocol):
    def next_back(self) -> Option[Item]: ...

    def rev(self) -> Rev[Item, P]:
        return Rev(self, par=self.par)


@dataclass(repr=True)
class Filter[Item, P: (True_, False_)](Iterator[Item, P]):
    iter: Iterator[Item, P]
    f: Fn1[Item, bool]
    par: P

    @override
    def next(self) -> Option[Item]:
        from serox.option import Null, Some

        match self.iter.next():
            case Some(v) if self.f(v):
                return Some(v)
            case Some(v):
                return self.next()
            case Null():
                return Null()


@dataclass(repr=True)
class FilterMap[Item, B, P: (True_, False_)](Iterator[B, P]):
    iter: Iterator[Item, P]
    f: Fn1[Item, Option[B]]
    par: P

    @override
    def next(self) -> Option[B]:
        from serox.option import Null, Some

        match self.iter.next():
            case Some(x):
                match y := self.f(x):
                    case Some(_):
                        return y
                    case Null():
                        return self.next()
            case Null():
                return Null()


@dataclass(repr=True)
class Map[Item, B, P: (True_, False_)](Iterator[B, P]):
    iter: Iterator[Item, P]
    f: Fn1[Item, B]
    par: P

    @override
    def next(self) -> Option[B]:
        return self.iter.next().map(self.f)


@dataclass(repr=True)
class Take[Item, P: (True_, False_)](Iterator[Item, P]):
    iter: Iterator[Item, P]
    _n: int
    par: P

    @override
    def next(self) -> Option[Item]:
        if self._n != 0:
            self._n -= 1
            return self.iter.next()
        else:
            from .option import Null

            return Null()

    @override
    def nth(self, n: int) -> Option[Item]:
        if self._n > n:
            self._n -= n + 1
            return self.iter.nth(n)
        else:
            if self._n > 0:
                _ = self.iter.nth(self.n - 1)
                self.n = 0
            return Null()


@dataclass(repr=True)
class TakeWhile[Item, P: (True_, False_)](Iterator[Item, P]):
    iter: Iterator[Item, P]
    predicate: Fn1[Item, bool]
    par: P
    flag: bool = False

    @override
    def next(self) -> Option[Item]:
        from serox.option import Null, Some

        if self.flag:
            return Null()
        else:
            match self.iter.next():
                case Null():
                    return Null()
                case Some(x) if self.predicate(x):
                    return Some(x)
                case _:
                    self.flag = True
                    return Null()


@dataclass(repr=True)
class Zip[A, B, P: (True_, False_)](Iterator[tuple[A, B], P]):
    a: Iterator[A, P]
    b: Iterator[B, P]
    par: P

    @override
    def next(self) -> Option[tuple[A, B]]:
        from serox.option import Null, Some

        match self.a.next():
            case Null():
                return Null()
            case Some(val_a):
                match self.b.next():
                    case Null():
                        return Null()
                    case Some(val_b):
                        return Some((val_a, val_b))


# Parametrising the first generic of `Iterator` as `Any` to avoid a circular import.
@dataclass(repr=True)
class ZipLongest[A, B, P: (True_, False_)](Iterator[Any, P]):
    a: Iterator[A, P]
    b: Iterator[B, P]
    par: P

    @override
    def next(self) -> Option[tuple[A, B] | tuple[Null[A], B] | tuple[A, Null[B]]]:
        from serox.option import Null, Some

        opt_a, opt_b = self.a.next(), self.b.next()
        match (opt_a, opt_b):
            case (Some(val_a), Some(val_b)):
                return Some((val_a, val_b))
            case (Null(), Some(val_b)):
                return Some((Null[A](), val_b))
            case (Some(val_a), Null()):
                return Some((val_a, Null[B]()))
            case _:
                return Null()


@dataclass(repr=True)
class Chain[A, P: (True_, False_)](Iterator[A, P]):
    a: Iterator[A, P]
    b: Iterator[A, P]
    par: P

    @override
    def next(self) -> Option[A]:
        from serox.option import Null, Some

        match item := self.a.next():
            case Null():
                return self.b.next()
            case Some(_):
                return item


@dataclass(repr=True)
class Rev[Item, P: (True_, False_)](Iterator[Item, P]):
    iter: DoubleEndedIterator[Item, P]
    par: P

    @override
    def next(self) -> Option[Item]:
        return self.iter.next_back()


class Extend[Item](Protocol):
    def extend(self, iter: Iterable[Item], /) -> None: ...

    def extend_one(self, item: Item) -> None:
        self.extend(Some(item))


@dataclass(repr=True, frozen=True, kw_only=True)
class Bridge[Item, Par: (True_, False_)](Iterator[Item, Par]):
    """
    A bridge between native Python iterators and `serox` ones.
    Can be parallel (`par = True`) or non-parallel (`par = False`).
    """

    iter: NativeIterator[Item]
    """The native Python iterator being bridged."""
    par: Par
    """Whether to parallelise the iterator."""

    @classmethod
    def new[Item2, Par2: (True_, False_)](
        cls, iter: NativeIterator[Item2], /, par: Par2 = False
    ) -> Bridge[Item2, Par2]:
        return Bridge(iter=iter, par=par)

    @classmethod
    def par_new[Item2](cls, iter: NativeIterator[Item2], /) -> Bridge[Item2, True_]:
        return Bridge(iter=iter, par=True)

    @override
    def next(self) -> Option[Item]:
        from .option import Null, Some

        try:
            return Some(self.iter.__next__())
        except StopIteration:
            return Null()


if TESTING:

    def test_par_invariance():
        from .collections import HashMap
        from .vec import Vec

        values = Vec(*range(4))
        keys = ["foo", "bar", "baz"]
        bridge = Bridge.new(iter(keys), par=False)
        bridge = Bridge(iter=iter(keys), par=False)
        mapped = values.iter().map(lambda x: x**2)
        _ = bridge.zip(mapped).collect(HashMap[str, int])

        bridge = Bridge.new(iter(keys), par=True)
        # shouldn't be able to combine parallel iterators with non-parallel ones
        # for consistent typing
        _ = bridge.zip(values.iter())  # pyright: ignore[reportArgumentType]

    def test_max():
        """Test the max method for iterators."""
        from .vec import Vec

        vec = Vec(1, 2, 3, 4, 5)
        assert vec.iter().max() == 5

    def test_iterator_methods():
        from .option import Null, Some
        from .result import Err, Ok
        from .vec import Vec

        vec = Vec(1, 2, 3, 4, 5)
        iter = vec.iter()

        # Test next
        assert iter.next() == Some(1)
        assert iter.next() == Some(2)

        # Test nth
        assert iter.nth(2) == Some(5)
        assert iter.nth(1) == Null()

        # Test advance_by
        iter = vec.iter()
        assert iter.advance_by(3) == Ok(None)
        assert iter.advance_by(3) == Err(1)

        # Test for_each
        result: list[int] = []
        iter = vec.iter()
        iter.for_each(lambda x: result.append(x))
        assert result == [1, 2, 3, 4, 5]

        # Test take
        iter = vec.iter().take(3)
        assert list(iter) == [1, 2, 3]

        # Test take_while
        iter = vec.iter().take_while(lambda x: x < 4)
        assert list(iter) == [1, 2, 3]

        # Test sum
        assert vec.iter().sum() == 15

        # Test product
        assert vec.iter().product() == 120

        # Test all
        assert vec.iter().all(lambda x: x > 0)
        assert not vec.iter().all(lambda x: x < 3)

        # Test any
        assert vec.iter().any(lambda x: x == 3)
        assert not vec.iter().any(lambda x: x == 6)

        # Test max
        assert vec.iter().max() == 5

        # Test min
        assert vec.iter().min() == 1

    def test_combinational_methods():
        from .option import Null, Some
        from .vec import Vec

        vec1 = Vec(1, 2, 3)
        vec2 = Vec(4, 5, 6)
        iter1 = vec1.iter()
        iter2 = vec2.iter()

        # Test zip
        zipped = iter1.zip(iter2)
        assert list(zipped) == [(1, 4), (2, 5), (3, 6)]

        # Test zip_longest
        iter1 = vec1.iter()
        iter2 = Vec(4, 5).iter()
        zipped_longest = iter1.zip_longest(iter2)
        assert list(zipped_longest) == [(1, 4), (2, 5), (3, Null())]

        # Test chain
        iter1 = vec1.iter()
        iter2 = vec2.iter()
        chained = iter1.chain(iter2)
        assert list(chained) == [1, 2, 3, 4, 5, 6]

        # Test map
        iter = vec1.iter().map(lambda x: x * 2)
        assert list(iter) == [2, 4, 6]

        # Test filter
        iter = vec1.iter().filter(lambda x: x % 2 == 1)
        assert list(iter) == [1, 3]

        # Test filter_map
        iter = vec1.iter().filter_map(lambda x: Some(x * 2) if x % 2 == 0 else Null())
        assert list(iter) == [4]

    def test_parallel_iterators():
        from .option import Null, Some
        from .result import Err, Ok
        from .vec import Vec

        vec = Vec(1, 2, 3, 4, 5)
        iter = vec.par_iter()

        # Test next
        assert iter.next() == Some(1)
        assert iter.next() == Some(2)

        # Test nth
        assert iter.nth(2) == Some(5)
        assert iter.nth(1) == Null()

        # Test advance_by
        iter = vec.par_iter()
        assert iter.advance_by(3) == Ok(None)
        assert iter.advance_by(3) == Err(1)

        # Test for_each
        result: list[int] = []
        iter = vec.par_iter()
        iter.for_each(lambda x: result.append(x))
        assert result == [1, 2, 3, 4, 5]

        # Test take
        iter = vec.par_iter().take(3)
        assert list(iter) == [1, 2, 3]

        # Test take_while
        iter = vec.par_iter().take_while(lambda x: x < 4)
        assert list(iter) == [1, 2, 3]

        # Test sum
        assert vec.par_iter().sum() == 15

        # Test product
        assert vec.par_iter().product() == 120

        # Test all
        assert vec.par_iter().all(lambda x: x > 0)
        assert not vec.par_iter().all(lambda x: x < 3)

        # Test any
        assert vec.par_iter().any(lambda x: x == 3)
        assert not vec.par_iter().any(lambda x: x == 6)

        # Test max
        assert vec.par_iter().max() == 5

        # Test min
        assert vec.par_iter().min() == 1



from typing import Any, TypeGuard  # noqa: I001


__all__ = [
    "default_if_none",
    "none",
    "some",
    "unwrap_or",
]


def some[T](value: T | None, /) -> TypeGuard[T]:
    """
    Returns ``True`` if the input is **not** ``None``
    (that is, if the ``Optional`` monad contains some value).

    :param value: Value to be checked.
    :returns: ``True`` if ``value`` is **not** ``None`` else ``False``.
    """
    return value is not None


def none(value: Any | None, /) -> TypeGuard[None]:
    """
    Returns ``True`` if the input **is** ``None``
    (that is, if the ``Optional`` monad contains no value).

    :param value: Value to be checked.
    :returns: ``True`` if ``value`` **is** ``None`` else ``False``.
    """
    return value is None


def unwrap_or[T](value: T | None, /, default: T) -> T:
    """
    Returns the input if the input is **not** None else the specified
    ``default`` value.

    :param value: Input to be unwrapped and returned if not ``None``.
    :param default: Default value to use if ``value`` is ``None``.
    :returns: ``default`` if ``value`` is ``None`` otherwise ``value``.
    """
    return default if value is None else value


default_if_none = unwrap_or



from typing import Protocol, runtime_checkable

__all__ = ["PathLike", "StrPath"]


@runtime_checkable
class PathLike[S: str | bytes](Protocol):
    def __fspath__(self) -> S: ...


type StrPath = str | PathLike[str]



from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Generator, Hashable, Iterable, Self, Sized, override
from typing import Iterator as NativeIterator

from serox.common import False_, True_
from serox.convert import Into
from serox.default import Default
from serox.iter import Extend, FromIterator, IntoIterator, IntoParIterator, Iterator
from serox.misc import Clone
from serox.option import Null, Option, Some

__all__ = [
    "HashSet",
]


class HashSet[T: Hashable](
    Default,
    Extend[T],
    FromIterator[T],
    IntoIterator[T],
    IntoParIterator[T],
    Into[set[T]],
    Clone,
    Sized,
):
    def __init__(self, *args: T) -> None:
        super().__init__()
        self.inner = set(args)

    @override
    @classmethod
    def default(cls) -> HashSet[T]:
        return HashSet[T]()

    @classmethod
    def new(cls) -> HashSet[T]:
        return HashSet[T].default()

    @override
    def into(self) -> set[T]:
        return self.inner

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({ repr(self.inner) })"

    def clear(self) -> None:
        self.inner.clear()

    def difference(self, other: Self) -> HashSet[T]:
        return HashSet[T](*(self.inner.difference(other.inner)))

    def __sub__(self, other: Self) -> HashSet[T]:
        return self.difference(other)

    def symmetric_difference(self, other: Self) -> HashSet[T]:
        return HashSet[T](*(self.inner.symmetric_difference(other.inner)))

    def union(self, other: Self) -> HashSet[T]:
        return HashSet[T](*(self.inner.union(other.inner)))

    def __or__(self, other: Self) -> HashSet[T]:
        return self.union(other)

    def intersection(self, other: Self) -> HashSet[T]:
        return HashSet[T](*(self.inner.intersection(other.inner)))

    def __and__(self, other: Self) -> HashSet[T]:
        return self.intersection(other)

    def __xor__(self, other: Self) -> HashSet[T]:
        return self.symmetric_difference(other)

    def contains(self, value: T) -> bool:
        return self.inner.__contains__(value)

    def is_disjoint(self, other: Self) -> bool:
        return self.inner.isdisjoint(other.inner)

    def is_subset(self, other: Self) -> bool:
        return self.inner.issubset(other.inner)

    def is_superset(self, other: Self) -> bool:
        return self.inner.issuperset(other.inner)

    def insert(self, value: T) -> bool:
        if preexisting := self.contains(value):
            self.inner.add(value)
        return preexisting

    @override
    def __eq__(self, other: Any) -> bool:
        match other:
            case HashSet():
                return self.inner == other.inner  # pyright: ignore[reportUnknownMemberType]
            case _:
                return False

    def len(self) -> int:
        return self.inner.__len__()

    @override
    def __len__(self) -> int:
        return self.len()

    def is_empty(self) -> bool:
        return self.len() == 0

    def take(self, value: T) -> Option[T]:
        if self.contains(value):
            self.remove(value)
            return Some(value)
        return Null()

    @override
    def extend(self, iter: Iterable[T], /) -> None:
        self.inner.update(iter)

    def remove(self, value: T, /) -> None:
        self.inner.remove(value)

    @override
    @classmethod
    def from_iter(cls, iter: Iterable[T], /) -> HashSet[T]:
        return HashSet(*iter)

    @override
    def iter(self) -> Iter[T, False_]:
        return Iter.new(self, par=False)

    def __iter__(self) -> Generator[T, None, None]:
        yield from self.iter()

    @override
    def par_iter(self) -> Iter[T, True_]:
        return Iter.new(self, par=True)

    @override
    def clone(self) -> HashSet[T]:
        return HashSet(*self.inner.copy())


@dataclass(repr=True, frozen=True, kw_only=True)
class Iter[Item, Par: (True_, False_)](Iterator[Item, Par]):
    iter: NativeIterator[Item]
    par: Par

    @classmethod
    def new[Item2, Par2: (True_, False_)](
        cls, data: HashSet[Item2], par: Par2 = True
    ) -> Iter[Item2, Par2]:
        iter_ = iter(data.inner)
        return Iter(iter=iter_, par=par)

    @override
    def next(self) -> Option[Item]:
        try:
            return Some(next(self.iter))
        except StopIteration:
            return Null()



from __future__ import annotations
import copy
from dataclasses import dataclass, field
from types import EllipsisType
from typing import Sized, override

from serox.conftest import TESTING
from serox.fmt import Debug
from serox.iter import DoubleEndedIterator
from serox.misc import Clone
from serox.option import Null, Option, Some

from .common import False_, True_

__all__ = ["Range"]

type Idx = int


@dataclass(
    eq=True,
    frozen=True,
    init=False,
    repr=False,
)
class Range[Par: (True_, False_)](
    DoubleEndedIterator[Idx, Par],
    Clone,
    Debug,
    Sized,
):
    __match_args__ = ("start", "end")

    start: Idx = field(init=False)
    end: Idx = field(init=False)
    par: Par = field(init=False)
    _ptr: Idx = field(init=False)

    def __init__(self, start: Idx | EllipsisType, end: Idx, *, par: Par = False) -> None:
        start = 0 if start is ... else start
        object.__setattr__(self, "start", start)
        object.__setattr__(self, "end", end)
        object.__setattr__(self, "par", par)
        object.__setattr__(self, "_ptr", start)
        super().__init__()

    @classmethod
    def new[Par2: (True_, False_)](
        cls, start: Idx | EllipsisType, end: Idx, *, par: Par2 = False
    ) -> Range[Par2]:
        return Range(start, end, par=par)

    @override
    def next(self) -> Option[Idx]:
        if self._ptr < self.end:
            item = Some(copy.copy(self._ptr))
            object.__setattr__(self, "_ptr", self._ptr + 1)
            return item
        return Null()

    @override
    def next_back(self) -> Option[Idx]:
        if self._ptr < self.end:
            item = Some(self.end - 1 - self._ptr)
            object.__setattr__(self, "_ptr", self._ptr + 1)
            return item
        return Null()

    @classmethod
    def from_sized[Par2: (True_, False_)](
        cls, sized: Sized, /, start: int = 0, *, par: Par2 = False
    ) -> Range[Par2]:
        return Range[Par2](start, len(sized), par=par)

    def contains(self, item: Idx, /) -> bool:
        return self.start <= item < self.end

    def is_empty(self) -> bool:
        return not (self.start < self.end)

    @override
    def clone(self) -> Range[Par]:
        return Range(start=self.start, end=self.end, par=self.par)

    def len(self) -> int:
        return max(self.end - self.start - 1, 0)

    @override
    def __len__(self) -> int:
        return self.len()

    @override
    def __repr__(self) -> str:
        return f"{self.start}..{self.end}"


if TESTING:
    from serox.vec import Vec

    def test_range():
        obj = Range(..., 10)
        _ = obj.par_bridge().map(lambda x: x**2).collect(Vec[int])
        for _ in obj:
            ...



from __future__ import annotations
import copy
from typing import Protocol, Self, override, runtime_checkable

__all__ = [
    "Addable",
    "Clone",
    "Clone",
    "Divisible",
    "Divisible_",
    "IdentityFunction",
    "IndexType",
    "Indexable",
    "Multiplicable",
    "Multiplicable_",
    "SelfAddable",
    "SelfMultiplicable",
    "Sized",
    "SizedIndexable",
    "SizedSelfAddable",
]


type IndexType = int


class Clone(Protocol):
    """
    A common protocol for the ability to duplicate an object.
    """

    def clone(self) -> Self:
        """
        Duplicate `self`.
        """
        return copy.deepcopy(self)


class Dupe(Protocol):
    """
    A cheap version of :class:`~Clone`.
    """

    def dupe(self) -> Self:
        """
        Cheaply duplicate `self`.
        """
        return copy.copy(self)


@runtime_checkable
class Sized(Protocol):
    def __len__(self) -> int: ...


@runtime_checkable
class Indexable[_IndexItem](Protocol):
    def __getitem__(self, index: IndexType, /) -> _IndexItem: ...


@runtime_checkable
class SizedIndexable[_IndexItem](Sized, Indexable[_IndexItem], Protocol): ...


@runtime_checkable
class Multiplicable[Other](Protocol):
    def __mul__(self, other: Other) -> Self: ...


@runtime_checkable
class Multiplicable_[Other](Protocol):
    def __imul__(self, other: Other) -> Self: ...


@runtime_checkable
class Divisible[Other](Protocol):
    def __div__(self, other: Other) -> Self: ...


@runtime_checkable
class Divisible_[Other](Protocol):
    def __idiv__(self, other: Other) -> Self: ...


@runtime_checkable
class Addable[Other](Protocol):
    def __add__(self, other: Other, /) -> Self: ...


@runtime_checkable
class RAddable[Other](Protocol):
    def __radd__(self, other: Other, /) -> Self: ...


@runtime_checkable
class LAddable[Other](Protocol):
    def __ladd__(self, other: Other, /) -> Self: ...


@runtime_checkable
class SelfAddable(Protocol):
    def __add__(self, other: Self, /) -> Self: ...


@runtime_checkable
class SelfMultiplicable(Protocol):
    def __mul__(self, other: Self, /) -> Self: ...


@runtime_checkable
class SizedSelfAddable(
    Sized,
    SelfAddable,
    Protocol,
):
    @override
    def __len__(self) -> int: ...

    @override
    def __add__(self, other: Self) -> Self: ...


@runtime_checkable
class IdentityFunction(Protocol):
    def __call__[T](self, __x: T) -> T: ...



# pyright: reportImportCycles=none
from __future__ import annotations
import copy
from dataclasses import dataclass
from typing import (
    Callable,
    Generator,
    Hashable,
    NoReturn,
    Protocol,
    TypeGuard,
    final,
    overload,
    override,
)

from serox.common import False_, True_
from serox.convert import From, Into
from serox.default import Default
from serox.iter import DoubleEndedIterator, IntoIterator
from serox.misc import Clone
from serox.result import Err, Ok, Result

from .conftest import TESTING

__all__ = [
    "Null",
    "Option",
    "Some",
    "is_null",
    "is_some",
]


@final
class UnwrapFailed(Exception):
    @override
    def __init__(self) -> None:
        super().__init__("Called `Option.unwrap()` on a Null value")


@final
class NullShortCircuit(Exception): ...


type Fn0[T] = Callable[[], T]
type Fn1[T, U] = Callable[[T], U]


class _Option[T](
    Clone,
    Default,
    Hashable,
    From[T | None],
    Into[T | None],
    IntoIterator[T],
    Protocol,
):
    @override
    def iter(self: Option[T]) -> Iter[T, False_]:
        """
        Returns an iterator over the possibly contained value.

        Examples
        ========
        .. code-block:: python
            x = Some(4)
            assert x.iter().next() == 4

            x: Option[int] = Null()
            assert x.iter().next() == Null()
        """
        return Iter(self, par=False)

    def __next__(self: Option[T]) -> Option[T]:
        return self.iter().next()

    def __iter__(self: Option[T]) -> Generator[T, None, None]:
        yield from Iter(self, par=False)

    def or_(self: Option[T], optb: Option[T], /) -> Option[T]:
        """
        Returns the option if it contains a value, otherwise returns `optb`.

        Arguments passed to or are eagerly evaluated; if you are passing the result of a function
        call, it is recommended to use :meth:`or_else`, which is lazily evaluated.
        """
        return self.__or__(optb)

    def __or__(self: Option[T], optb: Option[T], /) -> Option[T]:
        match x := self:
            case Some(_):
                return x
            case Null():
                return optb

    def or_else(self: Option[T], f: Fn0[Option[T]], /) -> Option[T]:
        """
        Returns the option if it contains a value, otherwise calls `f` and returns the result.
        """
        match x := self:
            case Some(_):
                return x
            case Null():
                return f()

    # Property emulative of Rust's `?` operator when
    # used in conjunction with [`serox::question_mark::qmark`].
    @property
    def q(self: Option[T]) -> T:
        """
        '?' operator for early exiting an `Option` returning function.

        Examples
        ========
        .. code-block:: python

            @qmark
            def some_function(value: Option[str]) -> Option[str]:
                return Some(value.q + "_suffix")
        """
        match self:
            case Some(x):
                return x
            case Null():
                raise NullShortCircuit

    def __invert__(self: Option[T]) -> T:
        return self.q

    @overload
    @classmethod
    def from_(cls: type[Some[T]], val: T, /) -> Some[T]: ...
    @overload
    @classmethod
    def from_(cls: type[Null[T]], val: None, /) -> Null[T]: ...
    @overload
    @classmethod
    def from_(cls: type[Some[T]], val: None, /) -> Null[T]: ...
    @overload
    @classmethod
    def from_(cls: type[Null[T]], val: T, /) -> Some[T]: ...
    @overload
    @classmethod
    def from_(cls: type[Option[T]], val: T | None, /) -> Option[T]: ...
    @override
    @classmethod
    def from_(cls: type[Option[T]], val: T | None, /) -> Option[T]:
        match val:
            case None:
                return Null[T]()
            case _:
                return Some(val)

    @override
    def into(self: Option[T]) -> T | None:
        match self:
            case Some(x):
                return x
            case Null():
                return None

    def is_some(self: Option[T]) -> bool:
        """
        Returns `True` if the option is a `Some` value.
        """
        return isinstance(self, Some)

    def is_null(self: Option[T]) -> bool:
        """
        Returns `True` if the option is a `Null` value.
        """
        return not self.is_some()

    @overload
    def unwrap(self: Some[T]) -> T: ...
    @overload
    def unwrap(self: Null[T]) -> NoReturn: ...
    def unwrap(self: Option[T]) -> T | NoReturn:
        """
        Returns the contained `Some` value, consuming the `self` value.

        Because this function may panic, its use is generally discouraged.
        Instead, prefer to use pattern matching and handle the `Null` case explicitly, or call
        :meth:`unwrap_or`, :meth:`unwrap_or_else`, or :meth:`unwrap_or_default`.

        raises :class:`UnwrapFailed`: if the `self` is value equals `Null`
        """
        match self:
            case Some(x):
                return x
            case Null():
                raise UnwrapFailed()

    def unwrap_or(self: Option[T], default: T, /) -> T:
        """
        Returns the contained `Some` value or a provided `default`.

        Arguments passed to :meth:`unwrap_or` are eagerly evaluated; if you are passing the result
        of a function call, it is recommended to use :meth:`unwrap_or_else`, which is lazily
        evaluated.

        Examples
        ========
        .. code-block:: python

            assert Some("car").unwrap_or("bike") == "car"
            assert Null[str].unwrap_or("bike") == "bike"
        """
        match self:
            case Some(x):
                return x
            case Null():
                return default

    def unwrap_or_else(self: Option[T], f: Callable[[], T], /) -> T:
        """
        Returns the contained `Some` value or computes it from a closure.

        Examples
        ========
        .. code-block:: python

            k = 10
            assert Some(4).unwrap_or_else(lambda: 2 * k) == 4
            assert Null[int]().unwrap_or_else(lambda: 2 * k) == 20
        """
        match self:
            case Some(x):
                return x
            case Null():
                return f()

    @overload
    def map[U](self: Null[T], f: Fn1[T, U], /) -> Null[U]: ...
    @overload
    def map[U](self: Some[T], f: Fn1[T, U], /) -> Some[U]: ...
    def map[U](self: Option[T], f: Fn1[T, U], /) -> Option[U]:
        """
        Maps an `Option[T]` to `Option[U]` by applying a function to a contained value (if `Some`)
        or returns `Null` (if `Null`).

        Examples
        ========
        .. code-block:: python

                maybe_some_string = Some("Hello, World!")
                maybe_some_len = maybe_some_string.map(lambda s: len(s))
                assert maybe_some_len == Some(13)

                x: Option[str] = Null()
                assert x.map(lambda s: len(s)) == Null()
        """
        match self:
            case Some(t):
                return Some(f(t))
            case Null():
                return Null[U]()

    def map_or[U](self: Option[T], default: U, f: Fn1[T, U], /) -> U:
        """
        Returns the provided default result (if `Null`), or applies a function to the contained
        value (if `Null`).

        Arguments passed to map_or are eagerly evaluated; if you are passing the result of a
        function call, it is recommended to use map_or_else, which is lazily evaluated.

        Examples
        ========
        .. code-block:: python

            x = Some("foo")
            assert x.map_or(42, lambda v: len(v)) == 3

            x: Option[str] = Null()
            assert x.map_or(42, lambda v: len(v)) == 42
        """
        match self:
            case Some(x):
                return f(x)
            case Null():
                return default

    def map_or_else[U](self: Option[T], default: Fn0[U], f: Fn1[T, U], /) -> U:
        """
        Computes a default function result (if `Null`), or applies a different function to the
        contained value (if `Some`).

        Examples
        ========
        .. code-block:: python

            k = 21
            x = Some("foo")
            assert x.map_or_else(lambda: 2 * k, lambda v: len(v)) == 3

            x: Option[str] = Null()
            assert x.map_or_else(lambda: 2 * k, lambda v: len(v)) == 42
        """
        match self:
            case Some(x):
                return f(x)
            case Null():
                return default()

    @overload
    def and_[U](self: Null[T], optb: Option[U], /) -> Null[U]: ...
    @overload
    def and_[U](self: Some[T], optb: Some[U], /) -> Some[U]: ...
    @overload
    def and_[U](self: Some[T], optb: Null[U], /) -> Null[U]: ...
    def and_[U](self: Option[T], optb: Option[U], /) -> Option[U]:
        """
        Returns `Null` if the option is `Null`, otherwise returns `optb`.

        Arguments passed to and are eagerly evaluated; if you are passing the result of a function
        call, it is recommended to use :meth:`and_then`, which is lazily evaluated.

        Examples
        ========

        .. code-block:: python

            x = Some(2)
            y: Option[str] = Null()
            assert x.and_(y) == Null()

            x: Option[int] = Null()
            y = Some("foo")
            assert x.and_(y) == Null()

            x = Some(2)
            y = Some("foo")
            assert x.and_(y) == Some("foo")

            x: Option[int] = Null()
            y: Option[str] = Null()
            assert x.and_(y) == Null()
        """
        match self:
            case Null():
                return Null[U]()
            case Some(_):
                return optb

    @overload
    def __and__[U](self: Null[T], optb: Option[U], /) -> Null[U]: ...
    @overload
    def __and__[U](self: Some[T], optb: Some[U], /) -> Some[U]: ...
    @overload
    def __and__[U](self: Some[T], optb: Null[U], /) -> Null[U]: ...
    def __and__[U](self: Option[T], optb: Option[U], /) -> Option[U]:
        return self.and_(optb)

    @overload
    def and_then[U](self: Null[T], f: Fn1[T, U], /) -> Null[U]: ...
    @overload
    def and_then[U](self: Some[T], f: Fn1[T, U], /) -> U: ...
    def and_then[U](self: Option[T], f: Fn1[T, U], /) -> U | Null[U]:
        """
        Returns `Null` if the option is `Null` otherwise calls `f` with the wrapped value and returns
        the result.

        Some languages call this operation 'flatmap'.

        Examples
        ========
        .. code-block:: python

            def sqrt_then_to_string(x: int) -> Option[str]:
                if x < 0:
                    return Null()
                return Some(str(x**0.5))


            assert Some(2).and_then(sq_then_to_string) == Some(str(4))
            assert Some(-1).and_then(sq_then_to_string) == Null()
            assert Null[str]().and_then(sq_then_to_string) == Null()

        Often used to chain fallible operations that may return `Null`.

        .. code-block:: python

            arr_2d = Vec(Vec("A0", "A1"), Vec("B0", "B1"))
            item_0_0 = arr_2d.first().and_then(lambda row: row.first())
            assert item_0_0 == Some("A1")

            item_2_0 = arr_2d.get(2).and_then(lambda row: row.get(0))
            assert item_2_0 == Null()
        """
        match self:
            case Some(x):
                return f(x)
            case Null():
                return Null[U]()

    def ok_or[E: Exception](self: Option[T], err: E, /) -> Result[T, E]:
        """
        Transforms the `Option[T]` into a `Result[T, E]`, mapping `Some(v)` to `Ok(v)` and `Null`
        to `Err(err)`.

        Arguments passed to :meth:`ok_or` are eagerly evaluated; if you are passing the result of a
        function call, it is recommended to use :meth:`ok_or_else`, which is lazily evaluated.

        Examples
        ========
        .. code-block:: python

            x = Some("foo")
            assert x.ok_or(0) == Ok("foo")

            x: Option[str] = Null()
            assert x.ok_or(0) == Err(0)
        """
        match self:
            case Some(x):
                return Ok(x)
            case Null():
                return Err(err)

    def ok_or_else[E: Exception](self: Option[T], err: Fn0[E], /) -> Result[T, E]:
        """
        Transforms the `Option[T]` into a `Result[T, E]`, mapping `Some(v)` to `Ok(v)` and `Null` to
        `Err(err())`.

        Examples
        =======
        .. code-block:: python

            x = Some("foo")
            assert x.ok_or_else(lambda: 0) == Ok("foo")

            x: Option[str] = Null()
            assert x.ok_or_else(lambda: 0) == Err(0)
        """
        match self:
            case Some(x):
                return Ok(x)
            case Null():
                return Err(err())

    def filter(self: Option[T], f: Fn1[T, bool]) -> Option[T]:
        """
        Returns `Null` if the option is `Null`, otherwise calls predicate with the wrapped value
        and returns:

        - `Some(t)` if predicate returns `True` (where `t` is the wrapped value), and
        - `None` if predicate returns `False`.

        This function works similar to :meth:`Iterator.filter()`.
        You can imagine the `Option[T]` being an iterator over one or zero elements.
        `filter()` lets you decide which elements to keep.
        """

        match self:
            case Some(x) if f(x):
                return Some(x)
            case _:
                return Null()

    @overload
    def clone(self: Null[T]) -> Null[T]: ...
    @overload
    def clone(self: Some[T]) -> Some[T]: ...
    @override
    def clone(self: Option[T]) -> Option[T]:
        match self:
            case Some(v):
                return Some(copy.deepcopy(v))
            case Null():
                return Null[T]()

    @override
    @classmethod
    def default(cls: type[Option[T]]) -> Option[T]:
        """
        Returns `Null`.
        """
        return Null()

    def zip[U](self: Option[T], other: Option[U]) -> Option[tuple[T, U]]:
        """
        Zips `self` with another `Option`.

        If `self` is `Some(s)` and `other` is `Some(o)`, this method returns `Some((s, o))`.
        Otherwise, `Null` is returned.

        Examples
        ========
        .. code-block:: python

            x = Some(1)
            y = Some("hi")
            z = Null[int]

            assert x.zip(y) == Some((1, "hi"))
            assert x.zip(z) == Null()
        """
        match (self, other):
            case (Some(a), Some(b)):
                return Some((a, b))
            case _:
                return Null()

    def transpose[E](self: Option[Result[T, E]]) -> Result[Option[T], E]:
        """
        Transposes an `Option` of a `Result` into a `Result` of an `Option`.

        `Null` will be mapped to `Ok(Null())`. `Some(Ok(_))` and `Some(Err(_))` will be mapped to
        `Ok(Some(_))` and `Err(_)`.

        Examples
        ========
        .. code-block:: python

            x = Result[Option[int], ValueError] = Ok(Some(5))
            y = Option[Result[int], ValueError] = Some(Ok(5))
            assert x == y.transpose()
        """
        from .result import Err, Ok

        match self:
            case Some(Ok(x)):
                return Ok(Some(x))
            case Some(Err(e)):
                return Err(e)
            case _:
                return Ok(Null())


@final
@dataclass(
    eq=True,
    frozen=True,
    repr=True,
)
class Null[T](_Option[T]):
    """
    No value.
    """

    __slots__ = ()

    # Emulate [`NoneType`].
    def __bool__(self) -> False_: ...


@final
@dataclass(
    eq=True,
    repr=True,
    slots=True,
    frozen=True,
)
class Some[T](_Option[T]):
    """
    Some value of type `T`.
    """

    __match_args__ = ("value",)
    value: T


type Option[T] = Some[T] | Null[T]
"""
Type `Option` represents an optional value: every `Option` is either `Some` and contains a value,
or `Null`, and does not. `Option` types are very common in Rust code, as they have a number of uses:

- Initial values
- Return values for functions that are not defined over their entire input range (partial functions)
- Return value for otherwise reporting simple errors, where `Null` is returned on error
- Optional class attributes
- Optional function arguments

Options are commonly paired with pattern matching to query the presence of a value and take action,
always accounting for the `Null` case.

.. code-block:: python

    def divide(numerator: float, denominator: float) -> Option[float]:
        if denominator == 0.0:
            return Null()
        else:
            return Some(numerator / denominator)

    # The return value of the function is an option
    result = divide(2.0, 3.0)

    # Pattern match to retrieve the value
    match result:
        #  The division was valid
        case Some(x):
            print(f"Result {x}")
        # The division was invalid
        case Null():
            print("Cannot divide by 0")
"""


def is_some[T](x: Option[T], /) -> TypeGuard[Some[T]]:
    """
    Returns `True` if the option is a `Some` value.

    Since `TypeGuard` does not allow for type-narrowing of the `self` parameter,
    this external function is needed to achieve that goal; :meth:`~Option.is_some`
    on its own does not suffice.
    """
    return x.is_some()


def is_null[T](x: Option[T], /) -> TypeGuard[Null[T]]:
    """
    Returns `True` if the option is a `Null` value.

    Since `TypeGuard` does not allow for type-narrowing of the `self` parameter,
    this external function is needed to achieve that goal; :meth:`~Option.is_null`
    on its own does not suffice.
    """
    return x.is_null()


@dataclass(
    repr=True,
    slots=True,
)
class Iter[Item, P: (True_, False_)](DoubleEndedIterator[Item, P]):
    item: Option[Item]
    par: P

    @override
    def next(self) -> Option[Item]:
        match self.item:
            case Some(x):
                self.item = Null()
                return Some(x)
            case Null():
                return Null()

    @override
    def next_back(self) -> Option[Item]:
        return self.item


if TESTING:
    from typing import cast

    import pytest

    @pytest.fixture()
    def foo_optional() -> str | None:
        return "foo"

    def test_from_into(foo_optional: str | None):
        opt = Some[str].from_(foo_optional)
        assert opt.unwrap() == "foo"
        opt_ls = cast(list[str] | None, ["bar", "baz"])
        opt = Some[list[str]].from_(opt_ls)
        assert opt.unwrap() == ["bar", "baz"]
        opt = Some[str].from_(None)
        with pytest.raises(UnwrapFailed):
            _ = opt.unwrap()

        deopted = opt.into()
        assert deopted is None

    @pytest.fixture()
    def foo_option() -> Option[str]:
        return Some("foo")

    def test_matching(foo_option: Option[str]):
        match foo_option:
            case Some(v):
                assert v == "foo"
            case Null():
                raise ValueError
        option2 = cast(Option[str], Null[str]())
        match option2:
            case Some(v):
                raise ValueError
            case v:
                ...

    def test_map(foo_option: Option[str]):
        res = foo_option.map(lambda x: x + "bar")
        assert res.unwrap() == "foobar"
        foo_option = Null[str]()
        res = foo_option.unwrap_or_else(lambda: "baz")
        assert res == "baz"

    def test_iter(foo_option: Option[str]):
        val = next(foo_option)
        assert val == Some("foo")
        val = list(iter(foo_option))
        assert len(val) == 1

        ls: list[str] = []
        ls.extend(foo_option)
        assert len(ls) == 1
        assert ls[0] == "foo"

        foo_option = Null[str]()
        val = list(iter(foo_option))
        assert len(val) == 0
